from launchable.utils.gzipgen import compress
import gzip
from unittest import TestCase

class GzippenTest(TestCase):
    def test_compress(self):
        """Basic sanity test of """
        encoded=b''.join(compress([b'Hello',b' ',b'world']))
        msg = gzip.decompress(encoded)
        print(msg)
        self.assertEqual(msg, b'Hello world')



from pathlib import Path
from unittest import mock
import responses  # type: ignore
import json
import gzip
from tests.cli_test_case import CliTestCase


class CypressTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/cypress/').resolve()

    @responses.activate
    def test_record_test_cypress(self):
        # test-result.xml was generated used to cypress-io/cypress-example-kitchensink
        # cypress run --reporter junit report.xml
        result = self.cli('record', 'tests',  '--session', self.session,
                          'cypress', str(self.test_files_dir) + "/test-result.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_subset_cypress(self):
        # test-report.xml is outputed from cypress/integration/examples/window.spec.js, so set it
        pipe = "cypress/integration/examples/window.spec.js"
        result = self.cli('subset', '--target', '10%',
                          '--session', self.session, 'cypress', input=pipe)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_empty_xml(self):
        # parse empty test report XML
        result = self.cli('record', 'tests',  '--session', self.session,
                          'cypress', str(self.test_files_dir) + "/empty.xml")
        self.assertEqual(result.exit_code, 0)
        self.assertIn(
            "close", responses.calls[1].request.url, "No record request")



import glob
import json
import os
import traceback
import click
from junitparser import JUnitXml, TestSuite, TestCase  # type: ignore
import xml.etree.ElementTree as ET
from typing import Callable, Dict, Generator, Iterator, List, Optional
from more_itertools import ichunked
from .case_event import CaseEvent
from ...utils.gzipgen import compress
from ...utils.http_client import LaunchableClient
from ...utils.token import parse_token
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.session import parse_session
from ...testpath import TestPathComponent
from ..helper import find_or_create_session
from http import HTTPStatus
from ...utils.click import KeyValueType
from ...utils.logger import Logger, AUDIT_LOG_FORMAT
import datetime
from dateutil.parser import parse


@click.group()
@click.option(
    '--base',
    'base_path',
    help='(Advanced) base directory to make test names portable',
    type=click.Path(exists=True, file_okay=False),
    metavar="DIR",
)
@click.option(
    '--session',
    'session',
    help='Test session ID',
    type=str,
)
@click.option(
    '--build',
    'build_name',
    help='build name',
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--debug',
    help='print request payload',
    default=False,
    is_flag=True,
)
@click.option(
    '--post-chunk',
    help='Post chunk',
    default=1000,
    type=int
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
@click.pass_context
def tests(context, base_path: str, session: Optional[str], build_name: Optional[str], debug: bool, post_chunk: int, flavor):
    session_id = find_or_create_session(context, session, build_name, flavor)

    token, org, workspace = parse_token()
    record_start_at = get_record_start_at(
        token, org, workspace, build_name, session)

    logger = Logger()

    # TODO: placed here to minimize invasion in this PR to reduce the likelihood of
    # PR merge hell. This should be moved to a top-level class

    class RecordTests:
        CaseEventType = Dict[str,str]
        # The most generic form of parsing, where a path to a test report
        # is turned into a generator by using CaseEvent.create()
        ParseFunc = Callable[[str], Generator[CaseEventType,None,None]]

        # A common mechanism to build ParseFunc by building JUnit XML report in-memory (or build it the usual way
        # and patch it to fix things up). This is handy as some libraries produce invalid / broken JUnit reports
        JUnitXmlParseFunc = Callable[[str],  ET.Element]

        @property
        def path_builder(self) -> CaseEvent.TestPathBuilder:
            """
            This function, if supplied, is used to build a test path
            that uniquely identifies a test case
            """
            return self._path_builder

        @path_builder.setter
        def path_builder(self, v: CaseEvent.TestPathBuilder):
            self._path_builder = v

        @property
        def parse_func(self) -> ParseFunc:
            return self._parse_func

        @parse_func.setter
        def parse_func(self, f: ParseFunc):
            self._parse_func = f

        # setter only property that sits on top of the parse_func property
        def set_junitxml_parse_func(self, f: JUnitXmlParseFunc):
            """
            Parse XML report file with the JUnit report file, possibly with the custom parser function 'f'
            that can be used to build JUnit ET.Element tree from scratch or do some patch up.

            If f=None, the default parse code from JUnitParser module is used.
            """
            def parse(report: str) -> Generator[RecordTests.CaseEventType,None,None]:
                # To understand JUnit XML format, https://llg.cubic.org/docs/junit/ is helpful
                # TODO: robustness: what's the best way to deal with broken XML file, if any?
                xml = JUnitXml.fromfile(report, f)
                if isinstance(xml, JUnitXml):
                    testsuites = [suite for suite in xml]
                elif isinstance(xml, TestSuite):
                    testsuites = [xml]
                else:
                    # TODO: what is a Pythonesque way to do this?
                    assert False

                for suite in testsuites:
                    for case in suite:
                        yield CaseEvent.from_case_and_suite(self.path_builder, case, suite, report)
            self.parse_func = parse
        junitxml_parse_func = property(None, set_junitxml_parse_func)


        @property
        def check_timestamp(self):
            return self._check_timestamp

        @check_timestamp.setter
        def check_timestamp(self, enable: bool):
            self._check_timestamp = enable

        def __init__(self):
            self.reports = []
            self.path_builder = CaseEvent.default_path_builder(base_path)
            self.junitxml_parse_func = None
            self.check_timestamp = True

        def make_file_path_component(self, filepath) -> TestPathComponent:
            """Create a single TestPathComponent from the given file path"""
            if base_path:
                filepath = os.path.relpath(filepath, start=base_path)
            return {"type": "file", "name": filepath}

        def report(self, junit_report_file: str):
            ctime = datetime.datetime.fromtimestamp(
                os.path.getctime(junit_report_file))

            if self.check_timestamp and ctime.timestamp() < record_start_at.timestamp():
                format = "%Y-%m-%d %H:%M:%S"
                logger.debug("skip: {} is old to report. start_record_at: {} file_created_at:{}".format(
                    junit_report_file, record_start_at.strftime(format), ctime.strftime(format)))
                return

            self.reports.append(junit_report_file)

        def scan(self, base: str, pattern: str):
            """
            Starting at the 'base' path, recursively add everything that matches the given GLOB pattern

            scan('build/test-reports', '**/*.xml')
            """
            for t in glob.iglob(os.path.join(base, pattern), recursive=True):
                self.report(t)

        def run(self):
            count = 0   # count number of test cases sent

            client = LaunchableClient(
                token, test_runner=context.invoked_subcommand)

            def testcases(reports: List[str]) -> Generator[RecordTests.CaseEventType,None,None]:
                exceptions = []
                for report in reports:
                    try:
                        yield from self.parse_func(report)

                    except Exception as e:
                        exceptions.append(Exception(
                            "Failed to process a report file: {}".format(report), e))

                if len(exceptions) > 0:
                    # defer XML parsing exceptions so that we can send what we can send before we bail out
                    raise Exception(exceptions)

            def splitter(iterable: Generator, size: int) -> Iterator[Iterator]:
                return ichunked(iterable, size)

            # generator that creates the payload incrementally
            def payload(cases: Generator[TestCase, None, None]) -> Generator[str, None, None]:
                nonlocal count
                yield '{"events": ['
                first = True        # used to control ',' in printing
                for case in cases:
                    if not first:
                        yield ','
                    first = False
                    count += 1
                    yield case
                yield ']}'

            def printer(f: Generator) -> Generator:
                for d in f:
                    print(d)
                    yield d

            def send(payload: Generator[TestCase, None, None]) -> None:
                # str -> bytes then gzip compress
                headers = {
                    "Content-Type": "application/json",
                    "Content-Encoding": "gzip",
                }

                def audit_log(f: Generator) -> Generator:
                    args = []
                    for d in f:
                        args.append(d)
                        yield d

                    logger.audit(AUDIT_LOG_FORMAT.format(
                        "post", "{}/events".format(session_id), headers, list(args)))

                payload = (s.encode() for s in audit_log(payload))
                payload = compress(payload)

                res = client.request(
                    "post", "{}/events".format(session_id), data=payload, headers=headers)

                if res.status_code == HTTPStatus.NOT_FOUND:
                    if session:
                        _, _, build, _ = parse_session(session)
                        click.echo(click.style(
                            "Session {} was not found. Make sure to run `launchable record session --build {}` before `launchable record tests`".format(session, build), 'yellow'), err=True)
                    elif build_name:
                        click.echo(click.style(
                            "Build {} was not found. Make sure to run `launchable record build --name {}` before `launchable record tests`".format(build_name, build_name), 'yellow'), err=True)

                res.raise_for_status()

            try:
                tc = (json.dumps(d) for d in testcases(self.reports))
                if debug:
                    tc = printer(tc)
                splitted_cases = splitter(tc, post_chunk)
                for chunk in splitted_cases:
                    send(payload(chunk))

                headers = {
                    "Content-Type": "application/json",
                    "Content-Encoding": "gzip",
                }
                res = client.request(
                    "patch", "{}/close".format(session_id), headers=headers)
                res.raise_for_status()

            except Exception as e:
                if os.getenv(REPORT_ERROR_KEY):
                    raise e
                else:
                    traceback.print_exc()
                    return

            click.echo("Recorded {} tests".format(count))
            if count == 0:
                click.echo(click.style(
                    "Looks like tests didn't run? If not, make sure the right files/directories are passed", 'yellow'))

    context.obj = RecordTests()


def get_record_start_at(token: str, org: str, workspace: str, build_name: Optional[str], session: Optional[str]):
    if session is None and build_name is None:
        raise click.UsageError(
            'Either --build or --session has to be specified')

    if session:
        _, _, build_name, _ = parse_session(session)

    client = LaunchableClient(token)
    headers = {
        "Content-Type": "application/json",
    }
    path = "/intake/organizations/{}/workspaces/{}/builds/{}".format(
        org, workspace, build_name)

    Logger().audit(AUDIT_LOG_FORMAT.format(
        "get", path, headers, None))

    res = client.request("get", path, headers=headers)
    if res.status_code == 404:
        click.echo(click.style(
            "Build {} was not found. Make sure to run `launchable record build --name {}` before `launchable record tests`".format(build_name, build_name), 'yellow'), err=True)
    elif res.status_code != 200:
        # to avoid stop report command
        return datetime.datetime.now()

    return parse_launchable_timeformat(res.json()["createdAt"])


def parse_launchable_timeformat(t: str) -> datetime.datetime:
    # e.g) "2021-04-01T09:35:47.934+00:00"
    try:
        return parse(t)
    except Exception as e:
        Logger().error(
            "parse time error {}. time: {}".format(str(e), t))
        return datetime.datetime.now()



import re

from . import launchable
from ..testpath import TestPath


def make_test_path(cls, case) -> TestPath:
    return [{'type': 'class', 'name': cls}, {'type': 'testcase', 'name': case}]


@launchable.subset
def subset(client):
    cls = ''
    for label in map(str.rstrip, client.stdin()):
        # handle Google Test's --gtest_list_tests output
        # FooTest.
        #  Bar
        #  Baz
        gtest_class = re.match(r'^([^\.]+)\.', label)
        if gtest_class:
            cls = gtest_class.group(1)
        gtest_case = re.match(r'^  ([^ ]+)', label)
        if gtest_case and cls:
            case = gtest_case.group(1)
            client.test_path(make_test_path(cls, case))

    client.formatter = lambda x: x[0]['name'] + "." + x[1]['name']
    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import click
import json
import os
import sys
from os.path import *
import glob
import gzip
from typing import Callable, Union, Optional
from ..utils.click import PERCENTAGE, DURATION
from ..utils.env_keys import REPORT_ERROR_KEY
from ..utils.http_client import LaunchableClient
from ..utils.token import parse_token
from ..testpath import TestPath
from ..utils.session import read_session
from .record.session import session as session_command
from .helper import find_or_create_session
from ..utils.click import KeyValueType
from ..utils.logger import Logger, AUDIT_LOG_FORMAT

# TODO: rename files and function accordingly once the PR landscape


@click.group(help="Subsetting tests")
@click.option(
    '--target',
    'target',
    help='subsetting target from 0% to 100%',
    type=PERCENTAGE,
)
@click.option(
    '--time',
    'duration',
    help='subsetting by absolute time, in seconds e.g) 300, 5m',
    type=DURATION,
)
@click.option(
    '--session',
    'session',
    help='Test session ID',
    type=str,
)
@click.option(
    '--base',
    'base_path',
    help='(Advanced) base directory to make test names portable',
    type=click.Path(exists=True, file_okay=False),
    metavar="DIR",
)
@click.option(
    '--build',
    'build_name',
    help='build name',
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--rest',
    'rest',
    help='output the rest of subset',
    type=str,
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,

)
@click.pass_context
def subset(context, target, session: Optional[str], base_path: Optional[str], build_name: Optional[str], rest: str, duration, flavor):
    token, org, workspace = parse_token()

    session_id = find_or_create_session(context, session, build_name, flavor)

    # TODO: placed here to minimize invasion in this PR to reduce the likelihood of
    # PR merge hell. This should be moved to a top-level class

    class Optimize:
        # test_paths: List[TestPath]  # doesn't work with Python 3.5

        # Where we take TestPath, we also accept a path name as a string.
        TestPathLike = Union[str, TestPath]

        def __init__(self):
            self.test_paths = []
            # TODO: robustness improvement.
            self._formatter = Optimize.default_formatter
            self._separator = "\n"

        @staticmethod
        def default_formatter(x: TestPath):
            """default formatter that's in line with to_test_path(str)"""
            file_name = x[0]['name']
            if base_path:
                # default behavior consistent with default_path_builder's relative path handling
                file_name = join(base_path, file_name)
            return file_name

        @property
        def formatter(self) -> Callable[[TestPath], str]:
            """
            This function, if supplied, is used to format test names
            from the format Launchable uses to the format test runners expect.
            """
            return self._formatter

        @formatter.setter
        def formatter(self, v: Callable[[TestPath], str]):
            self._formatter = v

        @property
        def separator(self) -> str:
            return self._separator

        @separator.setter
        def separator(self, s: str):
            self._separator = s

        def test_path(self, path: TestPathLike):
            def rel_base_path(path):
                if isinstance(path, str) and base_path:
                    return normpath(relpath(path, start=base_path))
                else:
                    return path

            if isinstance(path, str) and any(s in path for s in ('*', "?")):
                for i in glob.iglob(path, recursive=True):
                    if os.path.isfile(i):
                        self.test_paths.append(
                            self.to_test_path(rel_base_path(i)))
                return

            """register one test"""
            self.test_paths.append(self.to_test_path(rel_base_path(path)))

        def stdin(self):
            """
            Returns sys.stdin, but after ensuring that it's connected to something reasonable.

            This prevents a typical problem where users think CLI is hanging because
            they didn't feed anything from stdin
            """
            if sys.stdin.isatty():
                click.echo(click.style(
                    "Warning: this command reads from stdin but it doesn't appear to be connected to anything. Did you forget to pipe from another command?", fg='yellow'), err=True)
            return sys.stdin

        @staticmethod
        def to_test_path(x: TestPathLike) -> TestPath:
            """Convert input to a TestPath"""
            if isinstance(x, str):
                # default representation for a file
                return [{'type': 'file', 'name': x}]
            else:
                return x

        def scan(self, base: str, pattern: str, path_builder: Optional[Callable[[str], Union[TestPath, str, None]]] = None):
            """
            Starting at the 'base' path, recursively add everything that matches the given GLOB pattern

            scan('src/test/java', '**/*.java')

            'path_builder' is a function used to map file name into a custom test path.
            It takes a single string argument that represents the portion matched to the glob pattern,
            and its return value controls what happens to that file:
                - skip a file by returning a False-like object
                - if a str is returned, that's interpreted as a path name and
                  converted to the default test path representation. Typically, `os.path.join(base,file_name)
                - if a TestPath is returned, that's added as is
            """

            if path_builder == None:
                # default implementation of path_builder creates a file name relative to `source` so as not
                # to be affected by the path
                def default_path_builder(file_name):
                    file_name = join(base, file_name)
                    if base_path:
                        # relativize against `base_path` to make the path name portable
                        file_name = normpath(
                            relpath(file_name, start=base_path))
                    return file_name
                path_builder = default_path_builder

            for b in glob.iglob(base):
                for t in glob.iglob(join(b, pattern), recursive=True):
                    if path_builder:
                        path = path_builder(os.path.relpath(t, b))
                    if path:
                        self.test_paths.append(self.to_test_path(path))

        def get_payload(self, session_id, target, duration):
            payload = {
                "testPaths": self.test_paths,
                "session": {
                    # expecting just the last component, not the whole path
                    "id": os.path.basename(session_id)
                }
            }

            if target is not None:
                payload["target"] = target
            elif duration is not None:
                payload["goal"] = {
                    "type": "subset-by-absolute-time",
                    "duration": duration,
                }

            return payload

        def run(self):
            """called after tests are scanned to compute the optimized order"""

            # When Error occurs, return the test name as it is passed.
            output = self.test_paths

            if not session_id:
                # Session ID in --session is missing. It might be caused by Launchable API errors.
                pass
            else:
                try:
                    headers = {
                        "Content-Type": "application/json",
                        "Content-Encoding": "gzip",
                    }

                    payload = self.get_payload(session_id, target, duration)

                    path = "/intake/organizations/{}/workspaces/{}/subset".format(
                        org, workspace)

                    client = LaunchableClient(
                        token, test_runner=context.invoked_subcommand)

                    Logger().audit(AUDIT_LOG_FORMAT.format("post", path, headers, payload))

                    # temporarily extend the timeout because subset API response has become slow
                    # TODO: remove this line when API response return respose within 60 sec

                    timeout = (5, 180)
                    res = client.request("post", path, data=gzip.compress(json.dumps(
                        payload).encode()), headers=headers, timeout=timeout)

                    res.raise_for_status()
                    output = res.json()["testPaths"]
                except Exception as e:
                    if os.getenv(REPORT_ERROR_KEY):
                        raise e
                    else:
                        click.echo(e, err=True)
                    click.echo(click.style(
                        "Warning: the service failed to subset. Falling back to running all tests", fg='yellow'), err=True)

            if len(output) == 0:
                click.echo(click.style(
                    "Error: no tests found matching the path.", 'yellow'), err=True)
                return

            # regardless of whether we managed to talk to the service
            # we produce test names
            if rest is not None:
                rests = []

                subset = [self.formatter(t) for t in output]
                for test_path in self.test_paths:
                    p = self.formatter(test_path)
                    if p not in subset:
                        rests.append(p)

                if len(rests) == 0:
                    # no tests will be in the "rest" file. but add a test case to avoid failing tests using this
                    rests.append(subset[-1])

                open(rest, "w+").write(self.separator.join(rests))

            click.echo(self.separator.join(self.formatter(t)
                                           for t in output))

    context.obj = Optimize()



import unittest
import warnings


def ignore_warnings(test_func):
    def do_test(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            test_func(self, *args, **kwargs)
    return do_test



from unittest import TestCase
from launchable.commands.verify import *


class VersionTest(TestCase):
    def test_compare_version(self):
        def sign(x):
            if x<0:
                return -1
            if x>0:
                return 1
            return 0

        def f(expected, a,b):
            """Ensure symmetry on two sides"""
            self.assertEqual(sign(compare_version(a,b)), expected)
            self.assertEqual(sign(compare_version(b,a)), -expected)

        f(0, [1, 1, 0], [1, 1])     # 1.1.0 = 1.1
        f(1, [1, 1], [1, 0])        # 1.1 > 1.0
        f(1, [1, 0, 1], [1])        # 1.0.1 > 1

    def test_java_version(self):
        self.assertTrue(compare_java_version(
    """
    java version "1.8.0_144"
    Java(TM) SE Runtime Environment (build 1.8.0_144-b01)
    Java HotSpot(TM) 64-Bit Server VM (build 25.144-b01, mixed mode)
    """
        ) >= 0)

        self.assertTrue(compare_java_version(
    """
    java version "1.5.0_22"
    Java(TM) 2 Runtime Environment, Standard Edition (build 1.5.0_22-b03)
    Java HotSpot(TM) 64-Bit Server VM (build 1.5.0_22-b03, mixed mode)
    """
        ) < 0)



from pathlib import Path
import responses  # type: ignore
import json
import gzip
from launchable.utils.session import read_session
from tests.cli_test_case import CliTestCase


class GoTestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/go_test/').resolve()

    @responses.activate
    def test_subset_with_session(self):
        pipe = "TestExample1\nTestExample2\nTestExample3\nTestExample4\nok      github.com/launchableinc/rocket-car-gotest      0.268s"
        result = self.cli('subset', '--target', '10%',
                          '--session', self.session, 'go-test', input=pipe)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_subset_without_session(self):
        pipe = "TestExample1\nTestExample2\nTestExample3\nTestExample4\nok      github.com/launchableinc/rocket-car-gotest      0.268s"
        result = self.cli('subset', '--target', '10%', '--build',
                          self.build_name, 'go-test', input=pipe)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(
            responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_record_tests_with_session(self):
        result = self.cli('record', 'tests',  '--session',
                          self.session, 'go-test', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        self.assertIn(
            'events', responses.calls[1].request.url, 'call events API')
        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        # Remove timestamp because it depends on the machine clock
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

        self.assertIn(
            'close', responses.calls[2].request.url, 'call close API')

    @responses.activate
    def test_record_tests_without_session(self):
        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'go-test', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        self.assertEqual(read_session(self.build_name), self.session)

        self.assertIn(
            'events', responses.calls[2].request.url, 'call events API')
        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

        self.assertIn(
            'close', responses.calls[3].request.url, 'call close API')



from unittest import TestCase, mock
from launchable.utils.session import write_session, read_session, remove_session, clean_session_files, _get_session_id, parse_session
import os


class SessionTestClass(TestCase):
    build_name = '123'
    session_id = '/intake/organizations/launchableinc/workspaces/mothership/builds/123/test_sessions/13'

    def tearDown(self):
        clean_session_files()

    def test_write_read_remove(self):
        write_session(self.build_name, self.session_id)
        self.assertEqual(read_session(self.build_name), self.session_id)

        remove_session(self.build_name)
        self.assertEqual(read_session(self.build_name), None)

    def test_other_build(self):
        write_session(self.build_name, self.session_id)

        next_build_name = '124'
        next_session_id = '/intake/organizations/launchableinc/workspaces/mothership/builds/123/test_sessions/14'
        write_session(next_build_name, next_session_id)

        self.assertEqual(read_session(next_build_name), next_session_id)
        self.assertEqual(read_session(self.build_name), self.session_id)

    def test_read_before_write(self):
        self.assertEqual(read_session(self.build_name), None)

    def test_different_pid(self):
        # TODO
        pass

    def test_get_session_id(self):
        id = _get_session_id()

        with mock.patch.dict(os.environ, {"CIRCLECI": "TRUE", "CIRCLE_WORKFLOW_ID": "abc-123"}):
            id_in_circleci = _get_session_id()

            self.assertEqual(id_in_circleci, "abc-123")
            self.assertNotEqual(id, id_in_circleci)

    def test_parse_session(self):
        session = "/intake/organizations/org-name/workspaces/workspace-name/builds/build-name/test_sessions/123"
        org, workspace, build_name, session_id = parse_session(session)
        self.assertEqual(org, "org-name")
        self.assertEqual(workspace, "workspace-name")
        self.assertEqual(build_name, "build-name")
        self.assertEqual(session_id, "123")

        with self.assertRaises(Exception):
            parse_session("hoge/fuga")



from io import StringIO
import sys
from launchable.utils.logger import Logger
import launchable.utils.logger as logger
from unittest import TestCase
from unittest.mock import patch
import logging
import copy


class LoggerTest(TestCase):
    default_root_handlers = None

    def setUp(self):
        self.default_root_handlers = copy.copy(logging.root.handlers)

        # reset logging handler for test
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    def tearDown(self):
        logging.root.handlers = copy.copy(self.default_root_handlers)

    @patch("sys.stderr", new_callable=StringIO)
    def test_logging_default(self, mock_err):
        logging.basicConfig(level=logger.LOG_LEVEL_DEFAULT)
        l = Logger()
        l.audit("audit")
        l.info("info")
        l.warning("warn")
        l.debug("debug")
        self.assertEqual(mock_err.getvalue(), "")

    @patch("sys.stderr", new_callable=StringIO)
    def test_log_level_audit(self, mock_err):
        logging.basicConfig(level=logger.LOG_LEVEL_AUDIT)
        l = Logger()
        l.audit("audit")
        l.critical("critical")
        l.error("error")
        l.warning("warn")
        l.info("info")
        l.debug("debug")
        self.assertEqual(mock_err.getvalue(
        ), "AUDIT:launchable:audit\nCRITICAL:launchable:critical\nERROR:launchable:error\nWARNING:launchable:warn\n")



import os


def parse_token():
    try:
        token = os.environ["LAUNCHABLE_TOKEN"]
        _, user, _ = token.split(":", 2)
        org, workspace = user.split("/", 1)
    except ValueError:
        exit("Please set LAUNCHABLE_TOKEN environment variable to the Launchable API token")
    return token, org, workspace



from .cli_test_case import CliTestCase
from pathlib import Path


class PluginTest(CliTestCase):
    def test_plugin_loading(self):
        """
        Load plugins/foo.py as a plugin and execute its code
        """
        plugin_dir = Path(__file__).parent.joinpath('plugins').resolve()
        result = self.cli('--plugins', str(plugin_dir), 'record', 'tests',
                          '--session', '/intake/organizations/org/workspaces/work/builds/dummy/test_sessions/123', 'foo', 'alpha', 'bravo', 'charlie')
        self.assertTrue("foo:alpha" in result.stdout, result.stdout)
        self.assertTrue("foo:bravo" in result.stdout, result.stdout)
        self.assertTrue("foo:charlie" in result.stdout, result.stdout)



import os
from os.path import join
from pathlib import Path
import click
from junitparser import TestCase, TestSuite  # type: ignore
import json
from typing import List, Generator

from . import launchable
from ..testpath import TestPath
from ..utils.logger import Logger


def make_test_path(pkg, target) -> TestPath:
    return [{'type': 'package', 'name': pkg}, {'type': 'target', 'name': target}]


@launchable.subset
def subset(client):
    # Read targets from stdin, which generally looks like //foo/bar:zot
    for label in client.stdin():
        # //foo/bar:zot -> //foo/bar & zot
        if label.startswith('//'):
            pkg, target = label.rstrip('\n').split(':')
            # TODO: error checks and more robustness
            client.test_path(make_test_path(pkg.lstrip('//'), target))

    client.formatter = lambda x: x[0]['name'] + ":" + x[1]['name']
    client.run()


@click.argument('workspace', required=True)
@click.option('--build-event-json', 'build_event_json', help="set file path generated by --build_event_json_file", type=click.Path(exists=True), required=False)
@launchable.record.tests
def record_tests(client, workspace, build_event_json):
    """
    Takes Bazel workspace, then report all its test results
    """
    base = Path(workspace).joinpath('bazel-testlogs').resolve()
    if not base.exists():
        exit("No such directory: %s" % str(base))

    default_path_builder = client.path_builder

    def f(case: TestCase, suite: TestSuite, report_file: str) -> TestPath:
        # In Bazel, report path name contains package & target.
        # for example, for //foo/bar:zot, the report file is at bazel-testlogs/foo/bar/zot/test.xml
        # TODO: robustness
        pkgNtarget = report_file[len(str(base))+1:-len("/test.xml")]

        # last path component is the target, the rest is package
        # TODO: does this work correctly when on Windows?
        path = make_test_path(os.path.dirname(pkgNtarget),
                              os.path.basename(pkgNtarget))

        # let the normal path building kicks in
        path.extend(default_path_builder(case, suite, report_file))
        return path

    client.path_builder = f
    client.check_timestamp = False

    if build_event_json:
        for l in parse_build_event_json(build_event_json):
            if l is None:
                continue

            client.report(str(Path(base).joinpath(l, 'test.xml')))
    else:
        client.scan(str(base), '**/test.xml')

    client.run()


def parse_build_event_json(path: str) -> Generator:
    with open(path) as f:
        for line in f:
            try:
                d = json.loads(line)
            except Exception as e:
                Logger().error("Can not parse build event json {}".format(path))
                yield
            if "id" in d:
                if "testResult" in d["id"]:
                    if "label" in d["id"]["testResult"]:
                        label = d["id"]["testResult"]["label"]
                        # replace //foo/bar:zot to /foo/bar/zot
                        label = label.lstrip("/").replace(":", "/")
                        yield label



# Launchable's ingester Docker image with tag
# If you need up-to-date image, please replace the image tag
# You can see the docker image tags at
# https://hub.docker.com/repository/registry-1.docker.io/launchableinc/ingester/tags
ingester_image = (
    'launchableinc/ingester:'
    '976171f4005f8a1c6d0e706d20760890b44c5373'
)



import datetime
import os
import sys
from typing import Callable, Dict
from junitparser import Failure, Error, Skipped, TestCase, TestSuite  # type: ignore
from ...testpath import TestPath


class CaseEvent:
    EVENT_TYPE = "case"
    TEST_SKIPPED = 2
    TEST_PASSED = 1
    TEST_FAILED = 0

    # function that computes TestPath from a test case
    # The 3rd argument is the report file path
    TestPathBuilder = Callable[[TestCase, TestSuite, str], TestPath]

    @staticmethod
    def default_path_builder(base_path) -> TestPathBuilder:
        """
        Obtains a default TestPathBuilder that uses a base directory to relativize the file name
        """

        def f(case: TestCase, suite: TestSuite, report_file: str) -> TestPath:
            classname = case._elem.attrib.get("classname") or suite._elem.attrib.get("classname")
            filepath = case._elem.attrib.get(
                "file") or suite._elem.attrib.get("filepath")
            if filepath:
                filepath = os.path.relpath(filepath, start=base_path)

            test_path = []
            if filepath:
                test_path.append({"type": "file", "name": filepath})
            if classname:
                test_path.append({"type": "class", "name": classname})
            if case.name:
                test_path.append(
                    {"type": "testcase", "name": case._elem.attrib.get("name")})
            return test_path

        return f

    @classmethod
    def from_case_and_suite(cls, path_builder: TestPathBuilder, case: TestCase, suite: TestSuite, report_file: str, data: Dict = None) -> Dict:
        "Builds a JSON representation of CaseEvent from JUnitPaser objects"

        # TODO: reconsider the initial value of the status.
        status = CaseEvent.TEST_PASSED
        for r in case.result:
            if any(isinstance(r, s) for s in (Failure, Error)):
                status = CaseEvent.TEST_FAILED
                break
            elif isinstance(r, Skipped):
                status = CaseEvent.TEST_SKIPPED
        
        def path_canonicalizer(test_path: TestPath) -> TestPath:
            if sys.platform == "win32":
                for p in test_path:
                    p['name'] = p['name'].replace("\\", "/")
                
            return test_path

        return CaseEvent.create(
            path_canonicalizer(path_builder(case, suite, report_file)), case.time, status,
            case._elem.attrib.get("system-out"),
            case._elem.attrib.get("system-err"),
            suite.timestamp, data)

    @classmethod
    def create(cls, test_path: TestPath, duration_secs: float, status,
               stdout:str = None, stderr:str = None, timestamp: str = None, data: Dict = None) -> Dict:
        """
        Builds a JSON representation of CaseEvent from arbitrary set of values

        status:    TEST_FAILED or TEST_PASSED
        timestamp: ISO-8601 formatted date
        data:      arbitrary data to be submitted to the server. reserved for future enhancement.
        """
        return {
            "type": cls.EVENT_TYPE,
            "testPath": test_path,
            "duration": duration_secs,
            "status": status,
            "stdout": stdout or "",
            "stderr": stderr or "",
            "created_at": timestamp or datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "data": data
        }



import click
import os
import sys
from . import launchable


@click.argument('files', required=True, nargs=-1)
@launchable.subset
def subset(client, files):
    def parse(fname: str):
        '''
        Scan a file, directory full of *.rb, or @FILE
        '''
        if os.path.isdir(fname):
            client.scan(fname, '**/*_test.rb')
        elif fname == '@-':
            # read stdin
            for l in sys.stdin:
                parse(l)
        elif fname.startswith('@'):
            # read response file
            with open(fname[1:]) as f:
                for l in f:
                    parse(l)
        else:
            # assume it's a file
            client.test_path(fname)

    for f in files:
        parse(f)

    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import re
import click
import subprocess
import json
import os
from ...utils.token import parse_token
from .commit import commit
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.http_client import LaunchableClient
from ...utils.session import clean_session_files
from ...utils.logger import Logger, AUDIT_LOG_FORMAT


@click.command()
@click.option(
    '--name',
    'build_name',
    help='build name',
    required=True,
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--source',
    help='path to local Git workspace, optionally prefixed by a label.  '
    ' like --source path/to/ws or --source main=path/to/ws',
    default=["."],
    metavar="REPO_NAME",
    multiple=True
)
@click.option('--max-days',
    help="the maximum number of days to collect commits retroactively",
    default=30
)
@click.pass_context
def build(ctx, build_name, source, max_days):
    token, org, workspace = parse_token()

    clean_session_files(days_ago=14)

    # This command accepts REPO_NAME=REPO_DIST and REPO_DIST
    repos = [s.split('=') if re.match(r'[^=]+=[^=]+', s) else (s, s)
             for s in source]
    # TODO: if repo_dist is absolute path, warn the user that that's probably not what they want to do

    for (name, repo_dist) in repos:
        ctx.invoke(commit, source=repo_dist, max_days=max_days)

    sources = [(
        name,
        subprocess.check_output(
            "git rev-parse HEAD".split(), cwd=repo_dist
        ).decode().replace("\n", "")
    ) for name, repo_dist in repos]
    submodules = []
    for repo_name, repo_dist in repos:
        # invoke git directly because dulwich's submodule feature was broken
        submodule_stdouts = subprocess.check_output(
            "git submodule status --recursive".split(), cwd=repo_dist
        ).decode().splitlines()
        for submodule_stdout in submodule_stdouts:
            # the output is e.g.
            # "+bbf213437a65e82dd6dda4391ecc5d598200a6ce sub1 (heads/master)"
            matched = re.search(
                r"^[\+\-U ](?P<hash>[a-f0-9]{40}) (?P<name>\w+)",
                submodule_stdout
            )
            if matched:
                hash = matched.group('hash')
                name = matched.group('name')
                if hash and name:
                    submodules.append((repo_name+"/"+name, hash))

    # Note: currently becomes unique command args and submodules by the hash.
    # But they can be conflict between repositories.
    uniq_submodules = {hash: (name, hash)
                       for name, hash in sources + submodules}.values()

    try:
        commitHashes = [{
            'repositoryName': name,
            'commitHash': hash
        } for name, hash in uniq_submodules]

        if not (commitHashes[0]['repositoryName']
                and commitHashes[0]['commitHash']):
            exit('Please specify --source as --source .')

        payload = {
            "buildNumber": build_name,
            "commitHashes": commitHashes
        }

        headers = {
            "Content-Type": "application/json",
        }

        path = "/intake/organizations/{}/workspaces/{}/builds".format(
            org, workspace)

        client = LaunchableClient(token)

        Logger().audit(AUDIT_LOG_FORMAT.format("post", path, headers, payload))

        res = client.request("post", path, data=json.dumps(
            payload).encode(), headers=headers)
        res.raise_for_status()

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            print(e)



from tests.cli_test_case import CliTestCase
import responses  # type: ignore
import json


class SessionTest(CliTestCase):

    @responses.activate
    def test_run_session_without_flavor(self):
        result = self.cli("record", "session", "--build", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body)
        self.assert_json_orderless_equal({"flavors": {}}, payload)

    @responses.activate
    def test_run_session_with_flavor(self):
        result = self.cli("record", "session", "--build", self.build_name,
                          "--flavor", "key=value", "--flavor", "k", "v", "--flavor", "k e y = v a l u e")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body)
        self.assert_json_orderless_equal({"flavors": {
            "key": "value",
            "k": "v",
            "k e y": "v a l u e",
        }}, payload)

        with self.assertRaises(ValueError):
            result = self.cli("record", "session", "--build", self.build_name,
                              "--flavor", "only-key")
            self.assertEqual(result.exit_code, 1)
            self.assertTrue(
                "Expected key-value like --option kye=value or --option key value." in result.output)



from pathlib import Path
import responses  # type: ignore
import json
import gzip
import sys
from tests.cli_test_case import CliTestCase


class RspecTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/rspec/').resolve()
    result_file_path = test_files_dir.joinpath('record_test_result.json')

    @responses.activate
    def test_record_test_rspec(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          'rspec', str(self.test_files_dir.joinpath("rspec.xml")))
        self.assertEqual(result.exit_code, 0)
        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)



import click
from . import launchable
from xml.etree import ElementTree as ET


@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    for r in reports:
        client.report(r)

    def parse_func(p: str) -> ET.ElementTree:
        tree = ET.parse(p)
        for suites in tree.iter("testsuites"):
            if len(suites)==0:
                continue
            root_suite = suites.find('./testsuite[@name="Root Suite"]')
            if root_suite is not None:
                filepath = root_suite.get("file")
                if filepath is not None:
                    for suite in suites:
                        suite.attrib.update({"filepath": filepath})
        return tree

    client.junitxml_parse_func = parse_func
    client.run()


@launchable.subset
def subset(client):
    # read lines as test file names
    for t in client.stdin():
        client.test_path(t.rstrip("\n"))

    client.separator = ','
    client.run()



import os
import click
from . import launchable


@click.argument('source_roots', required=True, nargs=-1)
@launchable.subset
def subset(client, source_roots):
    def file2test(f: str):
        if f.endswith('.java') or f.endswith(".scala") or f.endswith(".kt"):
            f = f[:f.rindex('.')]   # remove extension
            # directory -> package name conversion
            cls_name = f.replace(os.path.sep, '.')
            return [{"type": "class", "name": cls_name}]
        else:
            return None

    for root in source_roots:
        client.scan(root, '**/*', file2test)

    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import os

import click

from . import launchable
from ..utils.file_name_pattern import jvm_test_pattern

@click.option('--bare',
    help='outputs class names alone',
    default=False,
    is_flag=True
)

@click.argument('source_roots', required=True, nargs=-1)
@launchable.subset
def subset(client, bare, source_roots):
    def file2test(f: str):
        if jvm_test_pattern.match(f):
            f = f[:f.rindex('.')]   # remove extension
            # directory -> package name conversion
            cls_name = f.replace(os.path.sep, '.')
            return [{"type": "class", "name": cls_name}]
        else:
            return None

    for root in source_roots:
        client.scan(root, '**/*', file2test)

    if bare:
        client.formatter = lambda x: x[0]['name']
    else:
        client.formatter = lambda x: "--tests {}".format(x[0]['name'])
        client.separator = ' '

    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import requests
import os
import platform
from launchable.version import __version__
from .logger import Logger

BASE_URL_KEY = "LAUNCHABLE_BASE_URL"
DEFAULT_BASE_URL = "https://api.mercury.launchableinc.com"


def get_base_url():
    return os.getenv(BASE_URL_KEY) or DEFAULT_BASE_URL


class LaunchableClient:
    def __init__(self, token: str, base_url: str = "", session: requests.Session = requests.Session(), test_runner: str = ""):
        self.base_url = base_url or get_base_url()
        self.session = session
        self.token = token
        self.test_runner = test_runner

    def request(self, method, path, **kwargs):
        headers = kwargs.pop("headers")
        url = self.base_url + path
        if 'timeout' not in kwargs:
            # (connection timeout, read timeout) in seconds
            kwargs['timeout'] = (5, 60)

        try:
            response = self.session.request(
                method, url, headers={**headers, **self._headers()}, **kwargs)
            Logger().debug(
                "received response status:{} message:{} headers:{}".format(
                    response.status_code, response.reason, response.headers)
            )
            return response
        except Exception as e:
            raise Exception("unable to post to %s" % url) from e

    def _headers(self):
        h = {
            "User-Agent": "Launchable/{} (Python {}, {})".format(__version__, platform.python_version(), platform.platform()),
            'Authorization': 'Bearer {}'.format(self.token),
        }

        if self.test_runner != "":
            h["launchable-test-runner"] = self.test_runner

        return h



import os
import sys
from pathlib import Path
import shutil
from typing import Optional
import datetime
import hashlib

SESSION_DIR_KEY = 'LAUNCHABLE_SESSION_DIR'
DEFAULT_SESSION_DIR = '~/.config/launchable/sessions/'


def _session_file_dir() -> Path:
    return Path(os.environ.get(SESSION_DIR_KEY) or DEFAULT_SESSION_DIR).expanduser()


def _session_file_path(build_name: str) -> Path:
    return _session_file_dir() / (hashlib.sha1("{}:{}".format(build_name, _get_session_id()).encode()).hexdigest() + ".txt")


def _get_session_id() -> str:
    if sys.platform == "win32":
        import wmi # type: ignore
        c = wmi.WMI()
        wql = "Associators of {{Win32_Process='{}'}} Where Resultclass = Win32_LogonSession Assocclass = Win32_SessionProcess".format(os.getpid())
        res = c.query(wql)
        id = res[0].LogonId
    else:
        id = str(os.getsid(os.getpid()))

    # CircleCI changes unix session id each steps, so set non change variable
    # https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
    if os.environ.get("CIRCLECI") is not None:
        id = os.environ.get("CIRCLE_WORKFLOW_ID") or ""

    return id


def read_session(build_name: str) -> Optional[str]:
    f = _session_file_path(build_name)
    try:
        if not f.exists():
            return None
        return f.read_text()
    except Exception as e:
        raise Exception("Can't read {}".format(f)) from e


def write_session(build_name: str, session_id: str) -> None:
    try:
        if not _session_file_dir().exists():
            _session_file_dir().mkdir(parents=True, exist_ok=True)

        _session_file_path(build_name).write_text(session_id)
    except Exception as e:
        raise Exception("Can't write to {}. Perhaps set the {} environment variable to specify an alternative writable path?".format(
            _session_file_path(build_name), SESSION_DIR_KEY)) from e


def remove_session(build_name: str) -> None:
    """
    Call it after closing a session
    """
    if _session_file_path(build_name).exists():
        _session_file_path(build_name).unlink()


def clean_session_files(days_ago: int = 0) -> None:
    """
    Call it each build start
    """
    if _session_file_dir().exists():
        for child in _session_file_dir().iterdir():
            file_created = datetime.datetime.fromtimestamp(
                child.stat().st_mtime)

            if sys.platform == "win32":
                # Windows sometimes misses to delete session files at Unit Test
                microseconds = -10
            else:
                microseconds = 0

            clean_date = datetime.datetime.now() - datetime.timedelta(days=days_ago, microseconds=microseconds)
            if file_created < clean_date:
                child.unlink()


def parse_session(session: str):
    try:
        # session format:
        # /intake/organizations/<org name>/workspaces/<workspace name>/builds/<build name>/test_sessions/<test session id>
        _, _, _, org, _, workspace, _, build_name, _, session_id = session.split(
            "/")
    except Exception as e:
        raise Exception("Can't parse session: {}".format(session)) from e

    return org, workspace, build_name, session_id



from .version import __version__
import click
import importlib, importlib.util
import logging
from os.path import dirname, basename, join
from glob import glob
from .commands.record import record
from .commands.subset import subset
from .commands.verify import verify
from .utils import logger


@click.group()
@click.version_option(version=__version__, prog_name='launchable-cli')
@click.option(
    '--log-level',
    'log_level',
    help='Set logger\'s log level (CRITICAL, ERROR, WARNING, AUDIT, INFO, DEBUG).',
    type=str,
    default=logger.LOG_LEVEL_DEFAULT_STR,
)
@click.option(
    '--plugins',
    'plugin_dir',
    help='Directory to load plugins from',
    type = click.Path(exists=True, file_okay=False)
)
def main(log_level, plugin_dir):
    level = logger.get_log_level(log_level)
    logging.basicConfig(level=level)

    # load all test runners
    for f in glob(join(dirname(__file__), 'test_runners', "*.py")):
        f = basename(f)[:-3]
        if f == '__init__':
            continue
        importlib.import_module('launchable.test_runners.%s' % f)

    # load all plugins
    if plugin_dir:
        for f in glob(join(plugin_dir, '*.py')):
            spec = importlib.util.spec_from_file_location("launchable.plugins.{}".format(basename(f)[:-3]), f)
            plugin = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(plugin)

main.add_command(record)
main.add_command(subset)
main.add_command(verify)


if __name__ == '__main__':
    main()



import os
from typing import List
import click
from . import launchable
from ..utils.file_name_pattern import jvm_test_pattern


@click.argument('source_roots', required=True, nargs=-1)
@launchable.subset
def subset(client, source_roots: List[str]):
    def file2test(f: str):
        if jvm_test_pattern.match(f):
            f = f[:f.rindex('.')]   # remove extension
            # directory -> package name conversion
            cls_name = f.replace(os.path.sep, '.')
            return [{"type": "class", "name": cls_name}]
        else:
            return None
    
    for root in source_roots:
        client.scan(root.rstrip('/'), "**/*Test.java", file2test)
    
    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import click
from typing import Optional
from .record.session import session as session_command
from ..utils.session import read_session


def _validate_session_and_build_name(session: Optional[str], build_name: Optional[str]):
    if session and build_name:
        raise click.UsageError(
            'Only one of --build or --session should be specified')

    if session is None and build_name is None:
        raise click.UsageError(
            'Either --build or --session has to be specified')


def find_or_create_session(context, session: Optional[str], build_name: Optional[str], flavor=[]) -> Optional[str]:
    _validate_session_and_build_name(session, build_name)

    if session:
        return session
    elif build_name:
        session_id = read_session(build_name)
        if session_id:
            return session_id
        else:
            context.invoke(
                session_command, build_name=build_name, save_session_file=True, print_session=False, flavor=flavor)
            return read_session(build_name)
    else:
        return None



from pathlib import Path
import responses  # type: ignore
import json
import gzip
from launchable.utils.session import read_session
from tests.cli_test_case import CliTestCase


class CTestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/ctest/').resolve()

    @responses.activate
    def test_subset_without_session(self):
        result = self.cli('subset', '--target', '10%', '--build',
                          self.build_name, 'ctest', str(self.test_files_dir.joinpath("ctest_list.json")))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    def test_record_test(self):
        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'ctest', str(self.test_files_dir) + "/Testing/**/Test.xml")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))

        for c in payload['events']:
            del c['created_at']
        self.assert_json_orderless_equal(payload, expected)



from pathlib import Path
import responses  # type: ignore
import json
import gzip
import sys
from pathlib import Path
from tests.cli_test_case import CliTestCase
from launchable.utils.http_client import get_base_url


class MinitestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/minitest/').resolve()
    result_file_path = test_files_dir.joinpath('record_test_result.json')

    @responses.activate
    def test_record_test_minitest(self):
        result = self.cli('record', 'tests',  '--session',
                          self.session, 'minitest', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())

        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_record_test_minitest_chunked(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          '--post-chunk', 5, 'minitest', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        payload1 = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        expected1 = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result_chunk1.json'))

        payload2 = json.loads(gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())
        expected2 = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result_chunk2.json'))

        # concat 1st and 2nd request for unstable order
        self.assert_json_orderless_equal(
            expected1['events'] + expected2['events'], payload1['events'] + payload2['events'])

    @responses.activate
    def test_subset(self):
        test_path = Path("test", "example_test.rb")
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
                          json={'testPaths': [[{'name': str(test_path)}]]}, status=200)

        result = self.cli('subset', '--target', '20%', '--session', self.session, '--base', str(self.test_files_dir),
                          'minitest', str(self.test_files_dir) + "/test/**/*.rb")

        self.assertEqual(result.exit_code, 0)
        output = Path(self.test_files_dir, "test", "example_test.rb")
        self.assertEqual(result.output.rstrip("\n"), str(output))

    @responses.activate
    def test_subset_with_invalid_path(self):
        result = self.cli('subset', '--target', '20%', '--session', self.session, '--base', str(self.test_files_dir),
                          'minitest', str(self.test_files_dir) + "/dummy")

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(
            "Error: no tests found matching the path." in result.output)



from typing import Dict, List

# Path component is a node in a tree.
# It's the equivalent of a short file/directory name in a file system.
# In our abstraction, it's represented as arbitrary bag of attributes
TestPathComponent = Dict[str, str]

# TestPath is a full path to a node in a tree from the root
# It's the equivalent of an absolute file name in a file system
TestPath = List[TestPathComponent]



from pathlib import Path
import responses  # type: ignore
import json
import gzip
from tests.cli_test_case import CliTestCase
from launchable.utils.http_client import get_base_url


class NUnitTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/nunit/').resolve()

    @responses.activate
    def test_subset(self):
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
                          json={'testPaths': [
                              [
                                  {"type": "Assembly", "name": "calc.dll"},
                                  {"type": "TestSuite", "name": "ParameterizedTests"},
                                  {"type": "TestFixture", "name": "MyTests"},
                                  {"type": "ParameterizedMethod", "name": "DivideTest"},
                                  {"type": "TestCase", "name": "DivideTest(12,3)"}
                              ],
                              [
                                  {"type": "Assembly", "name": "calc.dll"},
                                  {"type": "TestSuite", "name": "calc"},
                                  {"type": "TestFixture", "name": "Tests1"},
                                  {"type": "TestCase", "name": "Test1"}
                              ]
                          ]}, status=200)

        result = self.cli('subset', '--target', '10%', '--session', self.session,
                          'nunit', str(self.test_files_dir) + "/list.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))

        self.assert_json_orderless_equal(expected, payload)

        output = 'ParameterizedTests.MyTests.DivideTest(12,3)\ncalc.Tests1.Test1'
        self.assertEqual(result.output.rstrip('\n'), output)

    @responses.activate
    def test_record_test(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          'nunit', str(self.test_files_dir) + "/output.xml")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)



import click
import sys
import re

# click.Group has the notion of hidden commands but it doesn't allow us to easily add
# the same command under multiple names and hide all but one.


class GroupWithAlias(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super().__init__(name, commands, **attrs)
        self.aliases = {}

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, cmd_name) or self.aliases.get(cmd_name)

    def add_alias(self, name, cmd):
        self.aliases[name] = cmd


class PercentageType(click.ParamType):
    name = "percentage"

    def convert(self, value: str, param, ctx):
        try:
            if value.endswith('%'):
                x = float(value[:-1])/100
                if 0 <= x <= 100:
                    return x
        except ValueError:
            pass

        self.fail("Expected percentage like 50% but got '{}'".format(
            value), param, ctx)


class DurationType(click.ParamType):
    name = "duration"

    def convert(self, value: str, param, ctx):
        try:
            return convert_to_seconds(value)

        except ValueError:
            pass

        self.fail("Expected duration like 3600, 30m, 1h15m but got '{}'".format(
            value), param, ctx)


class KeyValueType(click.Option):
    error_message = "Expected key-value like --option kye=value or --option key value. but got '{}'"

    def __init__(self, *args, **kwargs):
        super(KeyValueType, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._key_value_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # case: --option key=value
            if '=' in value:
                kv = value.split('=')
                if len(kv) != 2:
                    raise ValueError(
                        self.error_message.format(value))

                value = tuple([kv[0].strip(), kv[1].strip()])
            # case: --option key value
            else:
                rargs = state.rargs
                # --option key-only
                if len(rargs) < 1:
                    raise ValueError(
                        self.error_message.format(value))
                # --option key --other-option / -option key - other-argument
                elif 0 < len(rargs) and any(rargs[0].startswith(p) for p in self._key_value_parser.prefixes):
                    raise ValueError(
                        self.error_message.format(" ".join([value, rargs[0]])))

                value = [value, state.rargs.pop(0)]

            self._previous_parser_process(tuple([value[0], value[1]]), state)

        retval = super(KeyValueType, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(
                name) or parser._short_opt.get(name)
            if our_parser:
                self._key_value_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break

        return retval


PERCENTAGE = PercentageType()
DURATION = DurationType()

# Can the output deal with Unicode emojis?
try:
    '\U0001f389'.encode(sys.stdout.encoding or "ascii")
    # If stdout encoding is unavailable, such as in case of pipe, err on the safe side (EMOJI=False)
    # This is a judgement call, but given that emojis do not serve functional purposes and purely decorative
    # erring on the safe side seems like a reasonable call.
    EMOJI = True
except UnicodeEncodeError as e:
    EMOJI = False

def emoji(s: str, fallback: str = ''):
    """
    Used to safely use Emoji where we can.

    Returns 's' in an environment where stdout can deal with emojis, but 'fallback' otherwise.
    """
    return s if EMOJI else fallback


def convert_to_seconds(s: str):
    units = {'s': 1, 'm': 60,
             'h': 60*60, 'd': 60*60*24, 'w': 60*60*24*7}

    if s.isdigit():
        return float(s)

    duration = 0
    for m in re.finditer(r'(?P<val>\d+)(?P<unit>[smhdw]?)', s, flags=re.I):
        val = m.group('val')
        unit = m.group('unit')

        if val is None or unit is None:
            raise ValueError("unable to parse: {}".format(s))

        u = units.get(unit)
        if u is None:
            raise ValueError("unable to parse: {}".format(s))

        duration += int(val) * u

    return float(duration)



import logging


LOG_LEVEL_DEFAULT = 100
LOG_LEVEL_DEFAULT_STR = "DEFAULT"
LOG_LEVEL_AUDIT = 25
LOG_LEVEL_AUDIT_STR = "AUDIT"

AUDIT_LOG_FORMAT = "send request method:{} path:{} headers:{} args:{}"

logging.addLevelName(LOG_LEVEL_AUDIT, "AUDIT")
logging.addLevelName(LOG_LEVEL_DEFAULT, "DEFAULT")


def get_log_level(level=str) -> int:
    level = level.lower()

    if level == "critical":
        return logging.CRITICAL
    elif level == "error":
        return logging.ERROR
    elif level == "warn" or level == "warning":
        return logging.WARNING
    elif level == "audit":
        return LOG_LEVEL_AUDIT
    elif level == "info":
        return logging.INFO
    elif level == "debug":
        return logging.DEBUG
    else:
        return LOG_LEVEL_DEFAULT


class Logger(object):
    def __init__(self, name="launchable"):
        logger = logging.getLogger(name)
        self.logger = logger

    def audit(self, msg, *args, **kargs):
        self.logger.log(LOG_LEVEL_AUDIT, msg, *args, **kargs)

    def debug(self, msg, *args, **kargs):
        self.logger.debug(msg, *args, **kargs)

    def info(self, msg, *args, **kargs):
        self.logger.info(msg, *args, **kargs)

    def warning(self, msg, *args, **kargs):
        self.logger.warning(msg, *args, **kargs)

    def error(self, msg, *args, **kargs):
        self.logger.error(msg, *args, **kargs)

    def critical(self, msg, *args, **kargs):
        self.logger.critical(msg, *args, **kargs)



from unittest import TestCase
from launchable.utils.http_client import LaunchableClient
import platform
from launchable.version import __version__


class LaunchableClientTest(TestCase):
    def test_header(self):
        cli = LaunchableClient("test", "/test")
        self.assertEqual(cli._headers(), {
            "Authorization": "Bearer test",
            "User-Agent": "Launchable/{} (Python {}, {})".format(__version__, platform.python_version(), platform.platform()),
        })

        cli = LaunchableClient("test", "/test", test_runner="dummy")
        self.assertEqual(cli._headers(), {
            "Authorization": "Bearer test",
            "User-Agent": "Launchable/{} (Python {}, {})".format(__version__, platform.python_version(), platform.platform()),
            "launchable-test-runner": "dummy"
        })



import click
import os
import json
from http import HTTPStatus
from ...utils.http_client import LaunchableClient
from ...utils.token import parse_token
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.session import write_session
from ...utils.click import KeyValueType
from ...utils.logger import Logger, AUDIT_LOG_FORMAT

LAUNCHABLE_SESSION_DIR_KEY = 'LAUNCHABLE_SESSION_DIR'


@click.command()
@click.option(
    '--build',
    'build_name',
    help='build name',
    required=True,
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--save-file/--no-save-file',
    'save_session_file',
    help='save session to file',
    default=True,
    metavar='SESSION_FILE'
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
def session(build_name: str, save_session_file: bool, print_session: bool = True, flavor=[]):
    """
    print_session is for barckward compatibility.
    If you run this `record session` standalone, the command should print the session ID because v1.1 users expect the beheivior. That is why the flag is default True.
    If you run this command from the other command such as `subset` and `record tests`, you should set print_session = False because users don't expect to print session ID to the subset output.
    """
    token, org, workspace = parse_token()
    headers = {
        "Content-Type": "application/json",
    }

    flavor_dict = {}
    for (k, v) in flavor:
        flavor_dict[k] = v

    client = LaunchableClient(token)
    try:
        session_path = "/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions".format(
            org, workspace, build_name)

        Logger().audit(AUDIT_LOG_FORMAT.format(
            "post", session_path, headers, {"flavors": flavor_dict}))

        res = client.request("post", session_path,
                             headers=headers, data=json.dumps(
                                 {"flavors": flavor_dict},
                             ))

        if res.status_code == HTTPStatus.NOT_FOUND:
            click.echo(click.style(
                "Build {} was not found. Make sure to run `launchable record build --name {}` before".format(build_name, build_name), 'yellow'), err=True)

        res.raise_for_status()
        session_id = res.json()['id']

        if save_session_file:
            write_session(build_name, "{}/{}".format(session_path, session_id))
        if print_session:
            # what we print here gets captured and passed to `--session` in later commands
            click.echo("{}/{}".format(session_path, session_id))

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(e, err=True)



from pathlib import Path
import responses  # type: ignore
import gzip
import sys
from tests.cli_test_case import CliTestCase
from launchable.commands.record.tests import parse_launchable_timeformat
import datetime


class TestsTest(CliTestCase):

    @responses.activate
    def test_filename_in_error_message(self):
        normal_xml = str(Path(__file__).parent.joinpath(
            '../../data/broken_xml/normal.xml').resolve())
        broken_xml = str(Path(__file__).parent.joinpath(
            '../../data/broken_xml/broken.xml').resolve())
        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'file', normal_xml, broken_xml)

        def remove_backslash(input: str) -> str:
            # Hack for Windowns. They containts double escaped backslash such as \\\\
            if sys.platform == "win32":
                return input.replace("\\", "")
            else:
                return input

        # making sure the offending file path name is being printed.
        self.assertIn(remove_backslash(broken_xml),
                      remove_backslash(result.output))

        # normal.xml
        self.assertIn('open_class_user_test.rb', gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())

    def test_parse_launchable_timeformat(self):
        t1 = "2021-04-01T09:35:47.934+00:00"  # 1617269747.934
        t2 = "2021-05-24T18:29:04.285+00:00"  # 1621880944.285
        t3 = "2021-05-32T26:29:04.285+00:00"  # invalid time format

        parse_launchable_time1 = parse_launchable_timeformat(t1)
        parse_launchable_time2 = parse_launchable_timeformat(t2)

        self.assertEqual(parse_launchable_time1.timestamp(), 1617269747.934)
        self.assertEqual(parse_launchable_time2.timestamp(), 1621880944.285)

        now = datetime.datetime.now()
        before = now + datetime.timedelta(seconds=-1)
        after = now + datetime.timedelta(seconds=1)

        self.assertTrue(before.timestamp() < parse_launchable_timeformat(
            t3).timestamp() < after.timestamp())



from xml.sax import *
from xml.sax.handler import ContentHandler
from typing import List, Dict, Callable
import sys
import re
import click

class Element:
    """Just like DOM element except it only knows about ancestors"""

    ## XML tag name
    # name: str

    ## parent element
    # parent: Element

    ## attributes. 'Attributes' class doesn't seem to exist
    # attrs: object

    ## tags captured at this context
    # tags: Dict[str,object]

    def __init__(self, parent: 'Element', name: str, attrs):
        self.name = name
        self.attrs = attrs
        self.parent = parent
        # start with a copy of parents, and we modify it with ours
        self.tags = parent.tags.copy() if parent else dict()    # type: Dict[str,object]

    def __str__(self):
        return "%s %s" % (self.name, self.tags)

class TagMatcher:
    """Matches to an attribute of an XML element and captures its value as a 'tag' """

    ## XML tag name to match
    # element: str
    ## XML attribute name to match
    # attr: str

    ## Name of the variable to capture
    # var: str

    def __init__(self, element: str, attr: str, var: str):
        self.element = element
        self.attr = attr
        self.var = var

    def matches(self, e: Element) -> str:
        return e.attrs.get(self.attr) if self.element==e.name else None

    @staticmethod
    def parse(spec :str) -> 'TagMatcher':
        """Parse a string like foo/@bar={zot}"""
        m = re.match(r"(\w+)/@(\w+)={(\w+)}", spec)
        if m:
            return TagMatcher(m.group(1), m.group(2), m.group(3))
        else:
            raise click.BadParameter("Invalid tag spec: %s"%spec)

class SaxParser(ContentHandler):
    """
    Parse XML in a streaming manner, capturing attribute values as specified by TagMatchers
    """

    ## represents the current element
    context = None  # type: Element

    # matchers: List[TagMatcher]

    # receiver: Callable[[Element],None]

    def __init__(self, matchers: List[TagMatcher], receiver: Callable[[Element],None]):
        super().__init__()
        self.matchers = matchers
        self.receiver = receiver

    def startElement(self, tag, attrs):
        self.context = Element(self.context, tag, attrs)

        # match tags at this element
        for m in self.matchers:
            v = m.matches(self.context)
            if v!=None:
                self.context.tags[m.var] = v

        # yield is more Pythonesque but because this is called from SAX parser
        # I'm assuming that won't work
        self.receiver(self.context)

    def endElement(self, tag):
        self.context = self.context.parent

    def parse(self, body):
        p = make_parser()
        p.setContentHandler(self)
        p.parse(body)


# Scaffold JUnit parser
# python -m launchable.utils.sax < result.xml
if __name__ == "__main__":
    def print_test_case(e: Element):
        if e.name=="testcase":
            print(e)

    SaxParser([
        TagMatcher.parse("testcase/@name={testcaseName}"),
        TagMatcher.parse("testsuite/@timestamp={timestamp}")
    ], print_test_case).parse(sys.stdin)



from pathlib import Path
import responses  # type: ignore
import json
import gzip
import itertools
from launchable.utils.session import read_session
from tests.cli_test_case import CliTestCase


class BazelTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/bazel/').resolve()
    subset_input = """Starting local Bazel server and connecting to it...
//src/test/java/com/ninjinkun:mylib_test9
//src/test/java/com/ninjinkun:mylib_test8
//src/test/java/com/ninjinkun:mylib_test7
//src/test/java/com/ninjinkun:mylib_test6
//src/test/java/com/ninjinkun:mylib_test5
//src/test/java/com/ninjinkun:mylib_test4
//src/test/java/com/ninjinkun:mylib_test3
//src/test/java/com/ninjinkun:mylib_test2
//src/test/java/com/ninjinkun:mylib_test1
Loading: 2 packages loaded
"""

    @responses.activate
    def test_subset(self):
        result = self.cli('subset', '--target', '10%',
                          '--build', self.build_name, 'bazel', input=self.subset_input)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(
            responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    def test_record_test(self):
        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'bazel', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))

        for c in payload['events']:
            del c['created_at']

        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    def test_record_test_with_build_event_json_file(self):
        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'bazel', '--build-event-json', str(self.test_files_dir.joinpath("build_event.json")), str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[2].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_with_build_event_json_result.json'))

        for c in payload['events']:
            del c['created_at']

        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    def test_subset_record_key_match(self):
        """
        Test recorded test results contain subset's test path
        """
        result = self.cli('subset', '--target', '10%',
                          '--build', self.build_name, 'bazel', input=self.subset_input)
        self.assertEqual(result.exit_code, 0)

        subset_payload = json.loads(gzip.decompress(
            responses.calls[1].request.body).decode())

        result = self.cli('record', 'tests', '--build',
                          self.build_name, 'bazel', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        record_payload = json.loads(gzip.decompress(
            b''.join(responses.calls[3].request.body)).decode())

        record_test_paths = itertools.chain.from_iterable(
            e['testPath'] for e in record_payload['events'])
        record_test_path_dict = {t['name']: t for t in record_test_paths}

        for test_paths in subset_payload['testPaths']:
            for subset_test_path in test_paths:
                record_test_path = record_test_path_dict.get(
                    subset_test_path['name'])
                self.assert_json_orderless_equal(
                    record_test_path, subset_test_path)



from pathlib import Path
from unittest import mock
import responses  # type: ignore
import json
import gzip
from tests.cli_test_case import CliTestCase
from launchable.utils.http_client import get_base_url
import tempfile
import os
from tests.helper import ignore_warnings


class GradleTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/gradle/').resolve()
    result_file_path = test_files_dir.joinpath('recursion/expected.json')

    @ignore_warnings
    @responses.activate
    def test_subset_without_session(self):
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
                          json={'testPaths': [[{'name': 'com.launchableinc.rocket_car_gradle.App2Test'}], [{'name': 'com.launchableinc.rocket_car_gradle.AppTest'}], [{'name': 'com.launchableinc.rocket_car_gradle.sub.App3Test'}], [{'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}]]}, status=200)
        result = self.cli('subset', '--target', '10%', '--build',
                          self.build_name, 'gradle', str(self.test_files_dir.joinpath('java/app/src/test').resolve()))
        # TODO: we need to assert on the request payload to make sure it found test list all right
        self.assertEqual(result.exit_code, 0)
        output = '--tests com.launchableinc.rocket_car_gradle.App2Test --tests com.launchableinc.rocket_car_gradle.AppTest --tests com.launchableinc.rocket_car_gradle.sub.App3Test --tests com.launchableinc.rocket_car_gradle.utils.UtilsTest'
        self.assertEqual(result.output.rstrip('\n'), output)

    @ignore_warnings
    @responses.activate
    def test_subset_rest(self):
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
                          json={'testPaths': [[{'type': 'class', 'name': 'com.launchableinc.rocket_car_gradle.App2Test'}], [{'type': 'class', 'name': 'com.launchableinc.rocket_car_gradle.AppTest'}], [{'type': 'class', 'name': 'com.launchableinc.rocket_car_gradle.utils.UtilsTest'}]]}, status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli('subset', '--target', '10%', '--build',
                          self.build_name, '--rest', rest.name, 'gradle', str(self.test_files_dir.joinpath('java/app/src/test/java').resolve()))

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output.rstrip(
            '\n'), "--tests com.launchableinc.rocket_car_gradle.App2Test --tests com.launchableinc.rocket_car_gradle.AppTest --tests com.launchableinc.rocket_car_gradle.utils.UtilsTest")
        self.assertEqual(rest.read().decode(
        ), '--tests com.launchableinc.rocket_car_gradle.sub.App3Test')
        rest.close()
        os.unlink(rest.name)

    @responses.activate
    def test_record_test_gradle(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          'gradle', str(self.test_files_dir) + "/**/reports")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())

        expected = self.load_json_from_file(self.result_file_path)
        self.assert_json_orderless_equal(expected, payload)



import click
from . import launchable
from xml.etree import ElementTree as ET
import os
from datetime import datetime
from junitparser import JUnitXml  # type: ignore
from ..testpath import TestPath


def parse_func(p: str) -> ET.ElementTree:
    datetime_format = '%Y%m%d %H:%M:%S.%f'

    original_tree = ET.parse(p)
    testsuite = ET.Element("testsuite", {"name": "robot"})
    for suite in original_tree.findall("suite/suite"):
        suite_name = suite.get('name')
        for test in suite.iter("test"):
            test_name = test.get('name')

            status_node = test.find('status')
            status = status_node.get(
                'status') if status_node is not None else None

            nested_status_node = test.find('./kw/status')
            nested_status = nested_status_node.get(
                'status') if nested_status_node is not None else None

            if status != None:
                start_time_str = status_node.get(
                    'starttime') if status_node is not None else ''
                end_time_str = status_node.get(
                    'endtime') if status_node is not None else ''

                if start_time_str != '' and end_time_str != '':
                    start_time = datetime.strptime(
                        str(start_time_str), datetime_format)
                    end_time = datetime.strptime(
                        str(end_time_str), datetime_format)

                    duration = end_time - start_time

                testcase = ET.SubElement(testsuite, "testcase", {
                    "name": str(test_name),
                    "classname": str(suite_name),
                    "time": str(duration.microseconds / 1000 / 1000) if duration is not None else '0',
                })

                if status == "FAIL":
                    failure = ET.SubElement(testcase, 'failure')

                    msg = test.find('kw/msg')
                    failure.text = msg.text if msg is not None else ''
                if status == "NOT_RUN" or nested_status == 'NOT_RUN':
                    skipped = ET.SubElement(testcase, "skipped")

    return ET.ElementTree(testsuite)


@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    for r in reports:
        client.report(r)

    client.junitxml_parse_func = parse_func
    client.run()


@click.argument('reports', required=True, nargs=-1)
@launchable.subset
def subset(client, reports):
    for r in reports:
        suite = JUnitXml.fromfile(r, parse_func)

        for case in suite:
            cls_name = case._elem.attrib.get("classname")
            name = case._elem.attrib.get('name')
            if cls_name != '' and name != '':
                client.test_path([{'type': 'class', 'name': cls_name}, {
                                 'type': 'testcase', 'name': name}])

    def formatter(x: TestPath):
        cls_name = ''
        case = ''
        for path in x:
            t = path['type']
            if t == 'class':
                cls_name = path['name']
            if t == 'testcase':
                case = path['name']

        if cls_name != '' and case != '':
            return "-s '{}' -t '{}'".format(
                cls_name, case)

        return ''

    client.formatter = formatter
    client.separator = " "
    client.run()



import os
import click
from urllib.parse import urlparse

from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.http_client import get_base_url
from ...utils.ingester_image import ingester_image
from ...utils.java import get_java_command

jar_file_path = os.path.normpath(os.path.join(
    os.path.dirname(__file__), "../../jar/exe_deploy.jar"))


@click.command()
@click.option('--source',
              help="repository path",
              default=os.getcwd(),
              type=click.Path(exists=True, file_okay=False),
              )
@click.option('--executable',
              help="collect commits with Jar or Docker",
              type=click.Choice(['jar', 'docker']),
              default='jar'
              )
@click.option('--max-days',
              help="the maximum number of days to collect commits retroactively",
              default=30
              )
def commit(source, executable, max_days):
    source = os.path.abspath(source)

    if executable == 'jar':
        try:
            exec_jar(source, max_days)
        except Exception as e:
            if os.getenv(REPORT_ERROR_KEY):
                raise e
            else:
                print(e)
    elif executable == 'docker':
        exec_docker(source, max_days)


def exec_jar(source, max_days):
    java = get_java_command()

    if not java:
        exit("You need to install Java or try --executable docker")

    base_url = get_base_url()

    https_proxy = os.getenv("HTTPS_PROXY")
    proxy_option = _build_proxy_option(https_proxy) if https_proxy else ""

    os.system(
        "{} {} -jar {} ingest:commit -endpoint {} -max-days {} {}"
        .format(java, proxy_option, jar_file_path, "{}/intake/".format(base_url), max_days, source))


def _build_proxy_option(https_proxy: str) -> str:
    if not (https_proxy.startswith("https://") or https_proxy.startswith("http://")):
        https_proxy = "https://" + https_proxy
    proxy_url = urlparse(https_proxy)

    options = []
    if proxy_url.hostname:
        options.append("-Dhttps.proxyHost={}".format(proxy_url.hostname))
    if proxy_url.port:
        options.append("-Dhttps.proxyPort={}".format(proxy_url.port))

    return "{} ".format(" ".join(options)) if len(options) else ""


def exec_docker(source, max_days):
    base_url = get_base_url()
    os.system(
        "docker run -u $(id -u) -i --rm "
        "-v {}:{} --env LAUNCHABLE_TOKEN {} ingest:commit -endpoint {} -max-days {} {}"
        .format(source, source, ingester_image, "{}/intake/".format(base_url), max_days, source)
    )



import click
from launchable.test_runners import launchable

@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    for r in reports:
        click.echo('foo:{}'.format(r))

@launchable.subset
def subset(client):
    click.echo("Subset!")




# MIT License, from https://github.com/leetreveil/gengzip
import time
import zlib
import struct
from builtins import int


def write32u(value):
    return struct.pack('<L', value)


def write_gzip_header():
    header = b''
    header += b'\037\213'  # magic header (0x1f, 0x8b)
    header += b'\010'  # compression method (deflate)
    header += b'\000'  # flags (not set)
    header += write32u(int(time.time()))  # file modification time
    header += b'\002'  # extra flags (maximum compression)
    header += b'\377'  # os type (unknown)
    return header


def write_gzip_footer(crc, size):
    footer = b''
    footer += write32u(crc)
    footer += write32u(size & 0xffffffff)
    return footer


def compress(d, compresslevel=6):
    """
    Takes a generator 'd' that provides streaem of data, then returns a wrapper generator
    that yields compressed gzip data stream
    """
    crc = zlib.crc32(b'') & 0xffffffff
    size = 0
    yield write_gzip_header()
    compress = zlib.compressobj(
        compresslevel, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
    for data in d:
        crc = zlib.crc32(data, crc) & 0xffffffff
        size += len(data)
        compressed = compress.compress(data)
        if len(compressed) > 0:
            yield compressed
    yield compress.flush() + write_gzip_footer(crc, size)



import os

import click
import platform, re, subprocess
from typing import List

from ..utils.env_keys import REPORT_ERROR_KEY
from ..utils.http_client import LaunchableClient
from ..utils.click import emoji
from ..utils.token import parse_token
from ..utils.java import get_java_command
from ..version import __version__ as version

def compare_version(a: List[int], b: List[int]):
    """Compare two version numbers represented as int arrays"""
    def pick(a,i):
        return a[i] if i<len(a) else 0

    for i in range(max(len(a),len(b))):
        d = pick(a,i)-pick(b,i)
        if d!=0:
            return d        # if they are different, we have the result
    return 0    # identical

def compare_java_version(output: str) -> int:
    """Check if the Java version meets what we need. returns >=0 if we meet the requirement"""
    for l in output.splitlines():
        if l.find("java version")!=-1:
            # l is like: java version "1.8.0_144"
            m = re.search('"([^"]+)"', l)
            if m:
                tokens = m.group(1).split(".")
                if len(tokens)>=2:
                    versions = [int(x) for x in tokens[0:2]]
                    required = [1,8]
                    return compare_version(versions,required)
    # couldn't determine, so err on the safe side
    return 0


def check_java_version(javacmd: str) -> int:
    """Check if the Java version meets what we need. returns >=0 if we meet the requirement"""
    v = subprocess.run([javacmd,"-version"], check=True, stderr=subprocess.PIPE, universal_newlines=True)
    return compare_java_version(v.stderr)

@click.command(name="verify")
def verify():
    try:
        if os.getenv("LAUNCHABLE_TOKEN") is None:
            raise click.UsageError(click.style("Could not find the LAUNCHABLE_TOKEN environment variable. Please check if you "
                                               "set it properly.", fg="red"))

        token, org, workspace = parse_token()

        click.echo("Organization: " + org)
        click.echo("Workspace: " + workspace)

        headers = {
            "Content-Type": "application/json",
        }

        path = "/intake/organizations/{}/workspaces/{}/verification".format(
            org, workspace)

        client = LaunchableClient(token)
        res = client.request("get", path, headers=headers)

        if res.status_code == 401:
            raise click.UsageError(click.style("Authentication failed. Most likely the value for the LAUNCHABLE_TOKEN "
                                               "environment variable is invalid.", fg="red"))

        res.raise_for_status()

        click.echo("Platform: " + platform.platform())
        click.echo("Python version: " + platform.python_version())

        java = get_java_command()

        if java is None:
            raise click.UsageError(click.style(
                "Java is not installed. Install Java version 8 or newer to use the Launchable CLI.", fg="red"))

        click.echo("Java command: " + java)
        click.echo("launchable version: " + version)

        # Level 2 check: versions. This is more fragile than just reporting the number, so we move
        # this out here

        if compare_version([int(x) for x in platform.python_version().split('.')], [3,5])<0:
            raise click.UsageError(click.style("Python 3.5 or later is required", fg="red"))

        if check_java_version(java) < 0:
            raise click.UsageError(click.style("Java 8 or later is required", fg="red"))

        click.echo(click.style(
            "Your CLI configuration is successfully verified"+emoji(" \U0001f389"), fg="green"))

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            print(e)
            import sys
            sys.exit(1)



from . import launchable

@launchable.subset
def subset(client):
    for case in client.stdin():
        # Avoid last line such as `ok      github.com/launchableinc/rocket-car-gotest      0.268s`
        if not ' ' in case:
            client.test_path([{'type': 'testcase', 'name': case.rstrip('\n')}])

    client.formatter = lambda x: "^{}$".format(x[0]['name'])
    client.separator = '|'
    client.run()


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import glob
import os

import click

from launchable.commands.record.case_event import CaseEvent
from launchable.utils.sax import SaxParser, Element
from . import launchable
from launchable.testpath import TestPath

# common code between 'subset' & 'record tests' to build up test path from nested <test-suite>s
def build_path(e: Element):
    pp = [] # type: TestPath
    if e.parent:
        pp = e.parent.tags.get('path') or []    # type: ignore
    if e.name == "test-suite":
        # <test-suite>s form a nested tree structure so capture those in path
        e.tags['path'] = pp + [{'type': e.attrs['type'], 'name': e.attrs['name']}]
    if e.name == "test-case":
        e.tags['path'] = pp + [{'type': 'TestCase', 'name': e.attrs['name']}]


@click.argument('report_xml', type=click.Path(exists=True), required=True)
@launchable.subset
def subset(client, report_xml):
    """
    Parse an XML file produced from NUnit --explore option to list up all the viable test cases
    """

    def on_element(e: Element):
        build_path(e)
        if e.name == "test-case":
            client.test_path(e.tags['path'])

    SaxParser([], on_element).parse(report_xml)

    # join all the names except when the type is ParameterizedMethod, because in that case test cases have
    # the name of the test method in it and ends up creating duplicates
    client.formatter = lambda x: '.'.join([c['name'] for c in x if c['type'] not in ['ParameterizedMethod','Assembly']])
    client.run()


@click.argument('report_xml', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, report_xml):
    # TODO: copied from ctest -- promote this pattern
    for root in report_xml:
        match = False
        for t in glob.iglob(root, recursive=True):
            match = True
            if os.path.isdir(t):
                client.scan(t, "*.xml")
            else:
                client.report(t)
        if not match:
            click.echo("No matches found: {}".format(root), err=True)

    def parse_func(report: str):
        events = []

        # parse <test-case> element into CaseEvent
        def on_element(e: Element):
            build_path(e)
            if e.name == "test-case":
                events.append(CaseEvent.create(
                    e.tags['path'], # type: ignore
                    float(e.attrs['duration']),
                    CaseEvent.TEST_PASSED if e.attrs['result'] == 'Passed' else CaseEvent.TEST_FAILED,
                    timestamp=e.attrs['start-time']))  # timestamp is already iso-8601 formatted

        SaxParser([], on_element).parse(report)

        # return the obtained events as a generator
        return (x for x in events)

    client.parse_func = parse_func
    client.run()



import click
from . import launchable
from xml.etree import ElementTree as ET
import os


@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    for r in reports:
        client.report(r)

    def parse_func(p: str) -> ET.ElementTree:
        tree = ET.parse(p)
        for suite in tree.iter("testsuite"):
            if len(suite) == 0:
                continue

            name = suite.get('name')
            if name is None:
                continue

            suite_name = name.split('.')
            if len(suite_name) < 2:
                continue

            file_name = suite_name[0] + ".feature"
            class_name = suite_name[1]
            suite.attrib.update({"filepath": file_name})

            for case in suite:
                case.attrib.update({"classname": class_name})

        return tree

    client.junitxml_parse_func = parse_func
    client.run()


@launchable.subset
def subset(client):
    for t in client.stdin():
        if 0 < t.find(".feature"):
            paths = os.path.split(t)
            if len(paths) < 2:
                continue

            file = paths[1]
            client.test_path(file.rstrip('\n'))

    client.separator = "|"
    client.run()



from unittest import TestCase

from launchable.utils.file_name_pattern import jvm_test_pattern

class FileNameHeuristicTest(TestCase):
  def test_jvm_file_name(self):
    file_names = [
      'LaunchableTest.java',
      'LaunchableTests.java',
      'LaunchableTestCase.java'
    ]
    for file_name in file_names:
      self.assertTrue(jvm_test_pattern.match(file_name))

    file_names = [
      'LaunchableTest2.java',
      'LaunchableTests2.java',
      'LaunchableTestCase2.java'
      'LaunchableTestUtil.java',
      'LaunchableTestss.java',
      'LaunchableTest2Case.java'
    ]
    for file_name in file_names:
      self.assertFalse(jvm_test_pattern.match(file_name))



from unittest import TestCase
from launchable.utils.click import convert_to_seconds


class DurationTypeTest(TestCase):
    def test_convert_to_seconds(self):
        self.assertEqual(convert_to_seconds('30s'), 30)
        self.assertEqual(convert_to_seconds('5m'), 300)
        self.assertEqual(convert_to_seconds('1h30m'), 5400)
        self.assertEqual(convert_to_seconds('1d10h15m'), 123300)
        self.assertEqual(convert_to_seconds('15m 1d 10h'), 123300)

        with self.assertRaises(ValueError):
            convert_to_seconds('1h30k')



import click, types, glob, os
from launchable.commands.subset import subset as subset_cmd
from launchable.commands.record.tests import tests as record_tests_cmd


def cmdname(m):
    """figure out the sub-command name from a test runner function"""

    # a.b.cde -> cde
    # xyz -> xyz
    #
    # In python module name the conventional separator is '_' but in command name,
    # it is '-', so we do replace that
    return m[m.rfind('.')+1:].replace('_','-')


def wrap(f, group, name=None):
    """
    Wraps a 'plugin' function into a click command and registers it to the given group.

    a plugin function receives the scanner object in its first argument
    """
    if not name:
        name = cmdname(f.__module__)
    d = click.command(name=name)
    cmd = d(click.pass_obj(f))
    group.add_command(cmd)
    return cmd


def subset(f): return wrap(f, subset_cmd)


record = types.SimpleNamespace()
record.tests = lambda f: wrap(f, record_tests_cmd)


class CommonRecordTestImpls:
    """
    Typical 'record tests' implementations that are reusable.
    """

    def __init__(self, module_name):
        self.cmdname = cmdname(module_name)

    def report_files(self):
        """
        Suitable for test runners that create a directory full of JUnit report files.

        'record tests' expect JUnit report/XML file names.
        """

        @click.argument('source_roots', required=True, nargs=-1)
        def record_tests(client, source_roots):
            # Accept both file names and GLOB patterns
            # Simple globs like '*.xml' can be dealt with by shell, but
            # not all shells consistently deal with advanced GLOBS like '**'
            # so it's worth supporting it here.
            for root in source_roots:
                match = False
                for t in glob.iglob(root, recursive=True):
                    match = True
                    if os.path.isdir(t):
                        client.scan(t, "*.xml")
                    else:
                        client.report(t)

                if not match:
                    # By following the shell convention, if the file doesn't exist or GLOB doesn't match anything,
                    # raise it as an error. Note this can happen for reasons other than a configuration error.
                    # For example, if a build catastrophically failed and no tests got run.
                    click.echo("No matches found: {}".format(root), err=True)
                    # intentionally exiting with zero
                    return

            client.run()

        return wrap(record_tests, record_tests_cmd, self.cmdname)



from pathlib import Path
import responses  # type: ignore
import json
import gzip
from tests.cli_test_case import CliTestCase


class GoogleTestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/googletest/').resolve()

    @responses.activate
    def test_subset(self):
        # I use "ctest -N" to get this list.
        pipe = """FooTest.
  Bar
  Baz
  Foo
        """
        result = self.cli('subset', '--target', '10%',
                          '--session', self.session, 'googletest', input=pipe)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_record_test_googletest(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          'googletest', str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('record_test_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_record_failed_test_googletest(self):
        # ./test_a --gtest_output=xml:output.xml
        result = self.cli('record', 'tests',  '--session', self.session,
                          'googletest', str(self.test_files_dir) + "/fail/")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())
        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('fail/record_test_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_record_empty_dir(self):
        path = 'latest/gtest_*_results.xml'
        result = self.cli('record', 'tests',  '--session', self.session,
                          'googletest', path)
        self.assertEqual(result.output.rstrip(
            '\n'), "No matches found: {}".format(path))
        self.assertEqual(result.exit_code, 0)



from click.testing import CliRunner
from launchable.__main__ import main
from launchable.version import __version__
from unittest import TestCase


class VersionTest(TestCase):
    def test_version(self):
        runner = CliRunner()
        result = runner.invoke(main, ['--version'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, 'launchable-cli, version {}\n'.format(__version__))



from pathlib import Path
from unittest import mock
import responses  # type: ignore
import json
import gzip
from tests.cli_test_case import CliTestCase
from launchable.utils.http_client import get_base_url


class MavenTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath(
        '../data/maven/').resolve()

    @responses.activate
    def test_subset(self):
        result = self.cli('subset', '--target', '10%', '--session',
                          self.session, 'maven', str(self.test_files_dir.joinpath('java/test/src/java/').resolve()))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    def test_subset_by_absolute_time(self):
        result = self.cli('subset', '--time', '1h30m', '--session',
                          self.session, 'maven', str(self.test_files_dir.joinpath('java/test/src/java/').resolve()))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            responses.calls[0].request.body).decode())

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath('subset_by_absolute_time_result.json'))

        self.assert_json_orderless_equal(expected, payload)

    @ responses.activate
    def test_record_test_maven(self):
        result = self.cli('record', 'tests',  '--session', self.session,
                          'maven', str(self.test_files_dir) + "/**/reports")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(
            b''.join(responses.calls[1].request.body)).decode())

        for e in payload["events"]:
            del e["created_at"]

        expected = self.load_json_from_file(
            self.test_files_dir.joinpath("record_test_result.json"))

        self.assert_json_orderless_equal(expected, payload)



import gzip
import json
import os
import unittest
import types
import responses  # type: ignore
import datetime

import click.testing
from click.testing import CliRunner

from launchable.__main__ import main
from launchable.utils.session import clean_session_files
from launchable.utils.http_client import get_base_url


class CliTestCase(unittest.TestCase):
    """
    Base class for setting up test to invoke CLI
    """
    organization = 'launchableinc'
    workspace = 'mothership'
    launchable_token = "v1:{}/{}:auth-token-sample".format(
        organization, workspace)
    session_id = 16
    build_name = 123
    session = "/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}".format(
        organization, workspace, build_name, session_id)

    def setUp(self):
        self.maxDiff = None
        os.environ['LAUNCHABLE_TOKEN'] = self.launchable_token

        responses.add(responses.POST, "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions".format(get_base_url(), self.organization, self.workspace, self.build_name),
                      json={'id': self.session_id}, status=200)
        responses.add(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(), self.organization, self.workspace),
                      json={'testPaths': []}, status=200)
        responses.add(responses.POST, "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}/events".format(get_base_url(), self.organization, self.workspace, self.build_name, self.session_id),
                      json={}, status=200)
        responses.add(responses.PATCH, "{}/intake/organizations/{}/workspaces/{}/builds/{}/test_sessions/{}/close".format(get_base_url(), self.organization, self.workspace, self.build_name, self.session_id),
                      json={}, status=200)
        responses.add(responses.GET, "{}/intake/organizations/{}/workspaces/{}/builds/{}".format(get_base_url(), self.organization, self.workspace, self.build_name),
                      json={'createdAt': "2020-01-02T03:45:56.123+00:00", 'id': 123}, status=200)

    def tearDown(self):
        clean_session_files()

    def cli(self, *args, **kwargs) -> click.testing.Result:
        """
        Invoke CLI command and returns its result
        """
        return CliRunner().invoke(main, args, catch_exceptions=False, **kwargs)

    def load_json_from_file(self, file):
        try:
            with file.open() as json_file:
                return json.load(json_file)
        except Exception as e:
            raise IOError("Failed to parse JSON {}".format(file)) from e

    def payload(self, mock_post):
        """
        Given a mock request object, capture the payload sent to the server
        """
        for (args, kwargs) in mock_post.call_args_list:
            if kwargs.get('data'):
                data = kwargs['data']
                if isinstance(data, types.GeneratorType):
                    data = b''.join(data)
                return data

        raise Exception("No data posted")

    def gzipped_json_payload(self, mock_post):
        return json.loads(gzip.decompress(self.payload(mock_post)).decode())

    def json_payload(self, mock_post):
        return json.loads(self.payload(mock_post))

    def assert_json_orderless_equal(self, a, b):
        """
        Compare two JSON trees ignoring orders of items in list & dict
        """
        def tree_sorted(obj):
            if isinstance(obj, dict):
                return sorted((k, tree_sorted(v)) for k, v in obj.items())
            if isinstance(obj, list):
                return sorted(tree_sorted(x) for x in obj)
            else:
                return obj

        self.assertEqual(tree_sorted(a), tree_sorted(b))



from .build import build
from .commit import commit
from .tests import tests
from .session import session
import click
from launchable.utils.click import GroupWithAlias


@click.group(cls=GroupWithAlias)
def record():
    pass


record.add_command(build)
record.add_command(commit)
record.add_command(tests)
# for backward compatibility
record.add_alias('test', tests)  # type: ignore
record.add_command(session)



import sys
import click
from xml.etree import ElementTree as ET
import json
from pathlib import Path
import glob
import os

from . import launchable
from ..testpath import TestPath

@click.argument('file', type=click.Path(exists=True))
@launchable.subset
def subset(client, file):
    if file:
        with Path(file).open() as json_file:
            data = json.load(json_file)
    else:
        data = json.loads(client.stdin())

    for test in data['tests']:
        case = test['name']
        client.test_path([{'type': 'testcase', 'name': case}])

    client.formatter = lambda x: "^{}$".format(x[0]['name'])
    client.seperator = '|'
    client.run()


@click.argument('source_roots', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, source_roots):
    for root in source_roots:
        match = False
        for t in glob.iglob(root, recursive=True):
            match = True
            if os.path.isdir(t):
                client.scan(t, "*.xml")
            else:
                client.report(t)
        if not match:
            click.echo("No matches found: {}".format(root), err=True)

    def parse_func(p: str) -> ET.ElementTree:
        """
        Convert from CTest own XML format to JUnit XML format
        The projections of these properties are based on
            https://github.com/rpavlik/jenkins-ctest-plugin/blob/master/ctest-to-junit.xsl
        """
        original_tree = ET.parse(p)

        testsuite = ET.Element("testsuite", {"name": "CTest"})
        test_count = 0
        failure_count = 0
        skip_count = 0

        for test in original_tree.findall("./Testing/Test"):
            test_name = test.find("Name")
            if test_name is not None:
                duration_node = test.find(
                    "./Results/NamedMeasurement[@name=\"Execution Time\"]/Value")
                measurement_node = test.find("Results/Measurement/Value")

                stdout = measurement_node.text if measurement_node is not None else ''
                duration = duration_node.text if duration_node  is not None else '0'

                testcase = ET.SubElement(testsuite, "testcase", {
                                        "name": test_name.text or '', "time": str(duration), "system-out": stdout or ''})

                system_out = ET.SubElement(testcase, "system-out")
                system_out.text = stdout

                test_count += 1
                status = test.get("Status")
                if status != None:
                    if status == "failed":
                        failure = ET.SubElement(testcase, "failure")
                        failure.text = stdout

                        failure_count += 1

                    if status == "notrun":
                        skipped = ET.SubElement(testcase, "skipped")
                        skipped.text = stdout

                        skip_count += 1

        testsuite.attrib.update({
                                "tests": str(test_count),
                                "time": "0",
                                "failures": str(failure_count),
                                "errors": "0",
                                "skipped": str(skip_count)
                                })

        return ET.ElementTree(testsuite)

    client.junitxml_parse_func = parse_func
    client.run()



from tests.cli_test_case import CliTestCase
from launchable.commands.record.commit import _build_proxy_option


class CommitTest(CliTestCase):
    def test_run_commit(self):
        """
        `record commit` command cause error because we can't mock Apache HTTP Client in Java
        """
        result = self.cli("record", "commit")
        self.assertEqual(result.exit_code, 0, "exit normally")

    def test_proxy_options(self):
        self.assertEqual(_build_proxy_option("https://some_proxy:1234"),
                         "-Dhttps.proxyHost=some_proxy -Dhttps.proxyPort=1234 ")
        self.assertEqual(_build_proxy_option("some_proxy:1234"),
                         "-Dhttps.proxyHost=some_proxy -Dhttps.proxyPort=1234 ")
        self.assertEqual(_build_proxy_option("some_proxy"),
                         "-Dhttps.proxyHost=some_proxy ")
        self.assertEqual(_build_proxy_option(
            "https://some_proxy"), "-Dhttps.proxyHost=some_proxy ")
        self.assertEqual(_build_proxy_option("http://yoyoyo"), 
                         "-Dhttps.proxyHost=yoyoyo ")



#
# The most bare-bone versions of the test runner support
#
import click
from junitparser import TestCase, TestSuite # type: ignore

from . import launchable
from ..testpath import TestPath


@launchable.subset
def subset(client):
    # read lines as test file names
    for t in client.stdin():
        client.test_path(t.rstrip("\n"))
    client.run()


@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    def path_builder(case: TestCase, suite: TestSuite, report_file: str) -> TestPath:
        """path builder that puts the file name first, which is consistent with the subset command"""
        def find_filename():
            """look for what looks like file names from test reports"""
            for e in [case, suite]:
                for a in ["file","filepath"]:
                    filepath = e._elem.attrib.get(a)
                    if filepath:
                        return filepath
            return None # failing to find a test name


        filepath = find_filename()
        if not filepath:
            raise click.ClickException("No file name found in %s" % report_file)

        # default test path in `subset` expects to have this file name
        test_path = [client.make_file_path_component(filepath)]
        if suite.name:
            test_path.append({"type": "testsuite", "name": suite.name})
        if case.name:
            test_path.append({"type": "testcase", "name": case.name})
        return test_path
    client.path_builder = path_builder

    for r in reports:
        client.report(r)
    client.run()



import os
import shutil

def get_java_command():
    if shutil.which("java"):
        return "java"

    if os.access(os.path.expandvars("$JAVA_HOME/bin/java"), os.X_OK):
        return os.path.expandvars("$JAVA_HOME/bin/java")

    return None



from unittest import TestCase
from launchable.commands.helper import _validate_session_and_build_name
import click


class TestHelper(TestCase):

    def test__validate_session_and_build_name(self):
        with self.assertRaises(click.UsageError):
            _validate_session_and_build_name(None, None)

        with self.assertRaises(click.UsageError):
            _validate_session_and_build_name("session", "build_name")

        _validate_session_and_build_name("session", None)
        _validate_session_and_build_name(None, "build_name")
